/*
 * Client library for speedfreqd
 *
 * (C) 2003 Jeremy Fitzhardinge <jeremy@goop.org>
 *
 * This software is licensed under the GNU GPL.  You know where you
 * can find a copy.
 */
#ifndef _SPEEDFREQ_H
#define _SPEEDFREQ_H

enum sf_policy {
	SF_NILPOLICY = -1,
	SF_POWERSAVE,
	SF_PERFORMANCE,
	SF_DYNAMIC,
	SF_FIXED,
};

struct sf_policyinfo {
	enum sf_policy policy;
	union {
		/* nothing for powersave */
		/* nothing for performance */
		struct sf_pol_dynamic {
			unsigned min, max;	/* frequency limits */
			unsigned current;	/* current dynamic speed */
		} dynamic;
		struct sf_pol_fixed {
			unsigned current;	/* current fixed speed (as set) */
		} fixed;
	};
};

struct sf_cpuinfo {
	unsigned min, max;	/* min and max cpu clock speeds */
	unsigned current;	/* current CPU clock (0 if unknown) */
	char	driver[32];	/* cpufreq driver name */
	float	idle;		/* idle ratio (0.0 == 100% busy; 1.0 == 100% idle) */
};

/*
 * State change monitoring callbacks.
 */
struct sf_monitorcb {
	void (*cpuinfo)(const struct sf_cpuinfo *);	/* callback for CPU state changes */
	void (*policy)(const struct sf_policyinfo *);	/* callback for policy changes */
};

/*
 * State change event structure.
 */
struct sf_mon_event {
	enum {
		EVT_INVALID = 0,		/* this structure contains junk */
		EVT_SHUTDOWN = 1,		/* server is going away */
		EVT_CPUINFO = 2,		/* this is a cpu state change event */
		EVT_POLICY = 3,			/* this is a policy change event */
	} type;
	union {
		struct sf_cpuinfo cpuinfo;	/* if .type == EVT_CPUINFO */
		struct sf_policyinfo policy;	/* if .type == EVT_POLICY */
	} u;
};

const char *sf_policyname(enum sf_policy);

/* Connect to server: returns 0 on failure */
int sf_connect(void);

/* Disconnect from server */
void sf_disconnect(void);

/* Get server version info (NULL on error).  Returns malloc'd string -
   you free it */
char *sf_getversion(void);

/* Get current policy info. */
int sf_getpolicy(struct sf_policyinfo *);

/* Policy setting functions: */

/* Power saving */
int sf_pol_powersave(void);

/* Performance */
int sf_pol_performance(void);

/* Dynamic.  Min and Max are the limits; specify 0 for CPU min and max */
int sf_pol_dynamic(unsigned min, unsigned max);

/* Fixed speed */
int sf_pol_fixed(unsigned set);

/* Get CPU's actual min, max and current speeds */
int sf_getcpuinfo(struct sf_cpuinfo *);

/* Tell server to shutdown */
int sf_shutdown(void);

/* Start monitoring client.  When called, it calls the callbacks to
   handle any pending monitoring events.  When it returns, it returns
   a file descriptor which can be select/polled for reading.  This
   should be called very regularly (every 100mS or so) to avoid losing
   events.

   Call sf_monitor() with a NULL callbacks pointer to end monitoring.

   If blocking is true, this call will block until there's some
   activity on the monitor connection.
*/
int sf_monitor(const struct sf_monitorcb *callbacks, int blocking);

/* Perform one step through the monitoring client, until either we
   receive an event or a timeout occurs.  If evt is NULL, shut down
   the monitoring client.  Otherwise, fill it out with the kind of
   event we received.

   Timeout is in milliseconds.  Zero means don't block; negative means
   block forever.  Call this function at least every 100ms to avoid
   missing events.

   Return value of -1 indicates an error; zero means nothing happened;
   1 means something meaningful occurred.
 */
int sf_poll(struct sf_mon_event *evt, int timeout);

/* Return the fd used to talk to the speedfreq daemon.
 */
int sf_get_mon_fd(void);

#endif /* _SPEEDFREQ_H */
