/*
 * (Quick hack in the shape of an) Automatic CPU speed governer.
 *
 * Measures the system's current idleness by reading /proc/uptime.  If
 * the system goes above 80% idle, the target CPU speed is dropped by
 * 100MHz.  If it goes below 30% idle, CPU speed is increased by
 * 100MHz.
 *
 * The daemon starts polling /proc/uptime every 300ms.  If the target
 * frequency is unchanged for 10 samples, the sample rate is dropped
 * to 1s.  It reverts to 300ms when there's a change.
 *
 * This now implements a client-server system.  If a server doesn't
 * already exist, speedfreq will become a server, listening to requests
 * on a socket.  Subsequent instances of speedfreq talk to the existing
 * server to do their work.  This allows users to change and query the
 * current policy.
 *
 * TODO:
 *   - have an adjustable ceiling for preferring power savings over performance
 *   - predicates for when to choose power savings:
 *     - running off battery
 *     - estimated running time remaining
 *     - thermal conditions
 *   - config file? command-line opts are easier
 *   - SMP support?
 *
 * Copyright 2003 Jeremy Fitzhardinge <jeremy@goop.org>
 *
 * This software is licensed under the GNU General Public License.
 * See COPYING for details.
 *
 * $Id: //depot/speedfreq/speedfreq.c#5 $
 */

#define _XOPEN_SOURCE 500	/* for pread() */
#define _BSD_SOURCE		/* for daemon() */

#include <stddef.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <sys/socket.h>
#include <sys/un.h>

#include "speedfreq.h"
#include "speedfreq-int.h"

extern const char version[];

/*
 * Tunables
 */

/* How often /proc/uptime is sampled */
#define SAMPLE_UNCHANGED	     10	/* number of unchanged samples before static */
#define SAMPLE_STATIC		1000000	/* steady state (uS) */
#define SAMPLE_DYNAMIC		 200000	/* on the move (uS) */
#define SAMPLE_JITTER		  10000	/* jitter sample to prevent beating */

/* Threshholds */
#define FREQ_UP			.3	/* fraction idle where we step up the clock */
#define FREQ_DOWN		.8	/* fraction idle where we step down the clock */
#define FREQ_STEP		100000	/* frequency step size (kHz) */

static const char *progname;
static const char *cpusys;

static int clo_debug;		/* debug messages */
static int clo_userchange;	/* users are allowed to change policy */

static int sockfd = -1;

static struct sf_cpuinfo cpuinfo;

#define POLICYLEN	64
static void policy_str(const struct sf_policyinfo *p, char *buf)
{
	switch(p->policy) {
	case SF_POWERSAVE:
	case SF_PERFORMANCE:
		sprintf(buf, "%s", sf_policyname(p->policy));
		break;
	case SF_FIXED:
		sprintf(buf, "fixed %d MHz", 
			p->fixed.current / 1000);
		break;
	case SF_DYNAMIC:
		sprintf(buf, "dynamic %d-%d MHz", 
			p->dynamic.min / 1000, p->dynamic.max / 1000);
		break;
	case SF_NILPOLICY:
		sprintf(buf, "none?");
		break;
	}	
}

/* Look for the cpufreq directory in /sys (it changes place from time
   to time) */
static int find_cpusys(void)
{
	static const char *paths[] = {
		"/sys/class/cpu/cpu%d/cpufreq",
		"/sys/devices/system/cpu/cpu%d/cpufreq",
		NULL
	};
	const char **p;
	char buf[128];
	struct stat st;

	for(p = paths; *p != NULL; p++) {
		sprintf(buf, *p, 0);
		if (stat(buf, &st) == 0 && S_ISDIR(st.st_mode))
			break;
	}

	cpusys = *p;

	return *p != NULL;
}

static int sysfile(const char *name, int cpu, unsigned mode)
{
	char buf[256];
	int len;
	int fd;

	len = snprintf(buf, sizeof(buf), cpusys, cpu);
	snprintf(buf+len, sizeof(buf)-len, "/%s", name);

	fd = open(buf, mode);
	if (fd == -1) {
		fprintf(stderr, "%s: can't open %s: %s\n",
			progname, buf, strerror(errno));
	}

	return fd;
}

static int readint(const char *name, unsigned *val)
{
	int fd;
	FILE *fp;
	int ret = 0;

	fd = sysfile(name, 0, O_RDONLY);

	if (fd == -1)
		return -1;

	fp = fdopen(fd, "r");

	if (fscanf(fp, "%u", val) != 1)
		ret = -1;

	fclose(fp);

	return ret;
}

static int readstr(const char *name, char *val, int vallen)
{
	int fd = sysfile(name, 0, O_RDONLY);
	int ret = 0;

	if (fd == -1)
		ret = -11;

	if (!ret && read(fd, val, vallen) == -1)
		ret = -11;

	close(fd);

	return ret;
}

static void writestr(const char *name, const char *val)
{
	int fd = sysfile(name, 0, O_WRONLY);

	write(fd, val, strlen(val));

	close(fd);
}

static void writeint(const char *name, unsigned val)
{
	int fd = sysfile(name, 0, O_WRONLY);
	FILE *fp = fdopen(fd, "w");

	fprintf(fp, "%u", val);

	fclose(fp);
}

static float getidle()
{
	static int uptime_fd = -1;
	static float ratio;
	static float tbase = 0., ibase = 0.;

	char buf[64];
	int len;
	float total, idle;

	if (uptime_fd == -1) {
		uptime_fd = open("/proc/uptime", O_RDONLY);
		if (uptime_fd == -1) {
			fprintf(stderr, "%s: can't open /proc/uptime: %s (/proc not mounted?)\n",
				progname, strerror(errno));
			exit(1);
		}
	}

	if ((len = pread(uptime_fd, buf, sizeof(buf)-1, 0)) == -1) {
		perror("pread");
		return ratio;
	}
	buf[len] = '\0';

	sscanf(buf, "%f %f", &total, &idle);

	if ((total-tbase) == 0.) /* no time passed at all, punt */
		return ratio;

	ratio = (idle-ibase)/(total-tbase);

	if (ratio < 0.)
		ratio = 0.;
	else if (ratio > 1.)
		ratio = 1.;

	tbase = total;
	ibase = idle;

	return ratio;
}

static struct monitor
{
	struct sockaddr_un addr;
	int addrlen;
	struct monitor *next;
} *monitors;

static int add_monitorclient(const struct sockaddr_un *addr, int addrlen)
{
	struct monitor *m = malloc(sizeof(*m));

	if (clo_debug)
		printf("%s: adding monitor\n", progname);

	if (m == NULL)
		return 0;

	memcpy(&m->addr, addr, addrlen);
	m->addrlen = addrlen;
	m->next = monitors;
	monitors = m;

	return 1;
}

static void del_monitorclient(const struct sockaddr_un *addr, int addrlen)
{
	struct monitor **prev = &monitors;
	struct monitor *m, *next;

	if (clo_debug)
		printf("%s: removing monitor(s)\n", progname);

	for(m = monitors; m != NULL; prev = &m->next, m = next) {
		next = m->next;

		if (addr == NULL ||
		    (m->addrlen == addrlen && memcmp(&m->addr, addr, addrlen) == 0)) {
			static const struct sfi_request cmd = { .cmd = SHUTDOWN };

			sendto(sockfd, &cmd, sizeof(cmd), MSG_NOSIGNAL,
			       (const struct sockaddr *)&m->addr, m->addrlen);

			free(m);
			m = NULL;
			*prev = next;
		} else
			prev = &m->next;
	}
}

static void sendnotify(const struct sockaddr_un *addr, int addrlen,
		       const struct sfi_request *cmd)
{
	struct monitor **prev = &monitors;
	struct monitor *m, *next;

	for(m = monitors; m != NULL; prev = &m->next, m = next) {
		next = m->next;

		if (addr == NULL || (m->addrlen == addrlen &&
				     memcmp(&m->addr, addr, addrlen) == 0)) {
			int ret = sendto(sockfd, cmd, sizeof(*cmd),
					 MSG_NOSIGNAL,
					 (const struct sockaddr *)&m->addr, m->addrlen);

			if (ret == -1 && ret != EAGAIN) {
				if (clo_debug)
					printf("removing monitor client with errno %s\n",
					       strerror(errno));
				free(m);
				m = NULL;
				*prev = next;
			} else
				prev = &m->next;
		} else
			prev = &m->next;
	}
}

static void notify_policychange(const struct sockaddr_un *addr, int addrlen,
				const struct sf_policyinfo *p)
{
	struct sfi_request cmd;

	memset(&cmd, 0, sizeof(cmd));

	if (clo_debug)
		printf("sending out policy change update\n");

	cmd.cmd = GETPOLICY;
	cmd.policy = *p;

	sendnotify(addr, addrlen, &cmd);
}

static void notify_cpuchange(const struct sockaddr_un *addr, int addrlen,
			     const struct sf_cpuinfo *c)
{
	struct sfi_request cmd;

	memset(&cmd, 0, sizeof(cmd));

	if (clo_debug)
		printf("sending out cpu change update: current=%dMHz\n", c->current / 1000);

	cmd.cmd = GETCPUINFO;
	cmd.cpuinfo = *c;

	sendnotify(addr, addrlen, &cmd);
}


typedef unsigned (*polfunc_t)(int, struct sf_policyinfo *p);

/* Very simple policy - tell the cpufreq driver to go into powersave,
   and we don't want to know anything more about it */
static unsigned pol_powersave(int init, struct sf_policyinfo *p)
{
	cpuinfo.idle = getidle();

	assert(p->policy == SF_POWERSAVE);
	if (init) {
		writestr("scaling_governor", "powersave");

		notify_policychange(NULL, 0, p);

		cpuinfo.current = cpuinfo.min;
		notify_cpuchange(NULL, 0, &cpuinfo);
	}

	return SAMPLE_STATIC;
}

/* Very simple policy - tell the cpufreq driver to go into
   performance, and we don't want to know anything more about it */
static unsigned pol_performance(int init, struct sf_policyinfo *p)
{
	cpuinfo.idle = getidle();

	assert(p->policy == SF_PERFORMANCE);
	if (init) {
		writestr("scaling_governor", "performance");

		notify_policychange(NULL, 0, p);

		cpuinfo.current = cpuinfo.max;
		notify_cpuchange(NULL, 0, &cpuinfo);
	}

	return SAMPLE_STATIC;
}

/* Set speed to a fixed clock */
static unsigned pol_fixed(int init, struct sf_policyinfo *p)
{
	cpuinfo.idle = getidle();

	assert(p->policy == SF_FIXED);
	if (init) {
		writestr("scaling_governor", "userspace");
		writeint("scaling_setspeed", p->fixed.current);

		notify_policychange(NULL, 0, p);

		if (p->fixed.current < cpuinfo.min)
			p->fixed.current = cpuinfo.min;
		if (p->fixed.current > cpuinfo.max)
			p->fixed.current = cpuinfo.max;
		
		readint("scaling_setspeed", &cpuinfo.current);
		notify_cpuchange(NULL, 0, &cpuinfo);
	}

	return SAMPLE_STATIC;
}

/* Dynamically change CPU speed, depending on how idle the CPU seems */
static unsigned pol_dynamic(int init, struct sf_policyinfo *p)
{
	static unsigned current, target;
	static unsigned sample = SAMPLE_DYNAMIC;
	static unsigned unchanged = 0;
	float ratio;

	assert(p->policy == SF_DYNAMIC);

	ratio = getidle();

	if (init) {
		writestr("scaling_governor", "userspace");

		readint("scaling_setspeed", &target);
		current = target;

		/* defaults */
		if (p->dynamic.min == 0)
			p->dynamic.min = cpuinfo.min;
		if (p->dynamic.max == 0)
			p->dynamic.max = cpuinfo.max;

		/* check CPU limits */
		if (p->dynamic.min < cpuinfo.min)
			p->dynamic.min = cpuinfo.min;
		if (p->dynamic.max > cpuinfo.max)
			p->dynamic.max = cpuinfo.max;

		/* sanity check */
		if (p->dynamic.min > p->dynamic.max)
			p->dynamic.min = p->dynamic.max;

		notify_policychange(NULL, 0, p);

		readint("scaling_setspeed", &cpuinfo.current);
		cpuinfo.idle = ratio;
		notify_cpuchange(NULL, 0, &cpuinfo);

		unchanged = 0;
	}

	if (ratio < FREQ_UP)
		target += FREQ_STEP;
	else if (ratio > FREQ_DOWN)
		target -= FREQ_STEP;

	if (target > p->dynamic.max)
		target = p->dynamic.max;
	else if (target < p->dynamic.min)
		target = p->dynamic.min;

	if (target != current) {
		if (clo_debug)
			printf("%g%% idle target=%dMHz\n",
			       ratio * 100, target/1000);

		writeint("scaling_setspeed", target);
		current = target;
		sample = SAMPLE_DYNAMIC;
		unchanged = 0;

		readint("scaling_setspeed", &cpuinfo.current);
		cpuinfo.idle = ratio;
		notify_cpuchange(NULL, 0, &cpuinfo);
	} else if (++unchanged > SAMPLE_UNCHANGED)
		sample = SAMPLE_STATIC;

	return sample + (random() % (SAMPLE_JITTER * 2)) - SAMPLE_JITTER;
}

static const struct policyinfo {
	polfunc_t	func;
} policies[] = {
	[SF_POWERSAVE]	= { pol_powersave },
	[SF_PERFORMANCE]= { pol_performance },
	[SF_DYNAMIC]	= { pol_dynamic },
	[SF_FIXED]	= { pol_fixed },
};

static polfunc_t getpolicyfunc(enum sf_policy p)
{
	assert(p < sizeof(policies)/sizeof(*policies));

	if (p == SF_NILPOLICY)
		return NULL;

	return policies[p].func;
}

#define H_NEWPOLICY	(1<<0)
#define H_QUIT		(1<<1)

static int recvcredfrom(int fd, void *data, int sz, int flags,
			struct sockaddr *addr, int *addrsz,
			struct ucred *ucp)
{
	struct msghdr msg = { 0, };
	struct iovec iov;
	char buf[CMSG_LEN(sizeof(struct ucred))];
	struct cmsghdr *cmsg;
	struct ucred *uc = NULL;
	int ret;

	ucp->uid = ucp->gid = ucp->pid = -1;

	msg.msg_name = addr;
	msg.msg_namelen = *addrsz;

	msg.msg_control = buf;
	msg.msg_controllen = sizeof(buf);

	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;

	iov.iov_base = data;
	iov.iov_len = sz;

	ret = recvmsg(fd, &msg, flags);

	if (ret == -1)
		return ret;

	for(cmsg = CMSG_FIRSTHDR(&msg);
	    cmsg != NULL;
	    cmsg = CMSG_NXTHDR(&msg, cmsg)) {
		if (cmsg->cmsg_level == SOL_SOCKET &&
		    cmsg->cmsg_type == SCM_CREDENTIALS) {
			uc = (struct ucred *)CMSG_DATA(cmsg);
			break;
		}
	}

	if (uc != NULL)
		*ucp = *uc;
	*addrsz = msg.msg_namelen;

	return ret;
}


/* Handle a command from the client */
static int handlecommand(struct sf_policyinfo *policy)
{
	int flags = 0;
	int donotify = 0;
	struct ucred uc;
	struct sockaddr_un addr;
	int addrlen = sizeof(addr);
	struct sfi_request cmd;

	if (recvcredfrom(sockfd, &cmd, sizeof(cmd), MSG_NOSIGNAL,
			 (struct sockaddr *)&addr, &addrlen,
			 &uc) <= 0)
		goto out;

	if (clo_debug) {
		printf("%s: got command %d\n", progname, cmd.cmd);
		printf("%s: got message from uid=%d gid=%d pid=%d\n",
		       progname,
		       uc.uid, uc.gid, uc.pid);
	}

	switch(cmd.cmd) {
	case GETVERSION:
		cmd.cmd = OK;
		strcpy(cmd.string, version);
		break;

	case GETCPUINFO:
		cmd.cmd = OK;
		cmd.cpuinfo = cpuinfo;
		break;

	case GETPOLICY:
		cmd.cmd = OK;
		cmd.policy = *policy;
		break;

	case SETPOLICY:
		if (clo_debug) {
			char buf[POLICYLEN];
			policy_str(&cmd.policy, buf);

			printf("%s: server setting policy to \"%s\"\n",
			       progname, buf);
		}

		if (!clo_userchange &&
		    (uc.uid != 0)) {
			cmd.cmd = ERROR;
			strcpy(cmd.string, "not allowed");
		} else {
			cmd.cmd = OK;

			*policy = cmd.policy;
			flags |= H_NEWPOLICY;
		}
		break;

	case SHUTDOWN:
		if (uc.uid != 0) {
			cmd.cmd = ERROR;
			strcpy(cmd.string, "not allowed");
		} else {
			cmd.cmd = OK;

			if (clo_debug)
				printf("%s: server exiting\n", progname);

			flags |= H_QUIT;
		}
		break;

	case MONITOR:
		if (cmd.enable) {
			if (add_monitorclient(&addr, addrlen)) {
				donotify = 1;
				cmd.cmd = OK;
			} else
				cmd.cmd = ERROR;
		} else {
			del_monitorclient(&addr, addrlen);
			cmd.cmd = OK;
		}
		break;

	default:
		if (clo_debug)
			printf("%s: bad command %d\n",
			       progname, cmd.cmd);
		cmd.cmd = ERROR;
		strcpy(cmd.string, "bad command");
		break;
	}

	if (sendto(sockfd, &cmd, sizeof(cmd), MSG_NOSIGNAL,
		   (const struct sockaddr *)&addr, addrlen) == -1) {
		perror("sendto");
		goto out;
	}

	if (donotify) {
		/* Urk.  Should make this more common. */
		notify_policychange(&addr, addrlen, policy);
		notify_cpuchange(&addr, addrlen, &cpuinfo);

		donotify = 0;
	}

  out:
	return flags;
}

/* Run the current policy while listening for commands from the
   client */
static int server(struct sf_policyinfo *policy)
{
	struct timeval tv;
	fd_set readset;
	int quit = 0;
	int newpolicy = 0;
	polfunc_t polfunc = getpolicyfunc(policy->policy);
	unsigned timeout = (*polfunc)(1, policy);

	memset(&readset, 0, sizeof(readset));

	while (!quit) {
		int ret;

		tv.tv_sec = timeout / 1000000;
		tv.tv_usec = timeout % 1000000;
		FD_SET(sockfd, &readset);

		ret = select(sockfd+1, &readset, NULL, NULL, timeout == 0 ? NULL : &tv);

		if (ret == -1) {
			perror("select");
			return 1;
		}

		if (ret == 1) {
			ret = handlecommand(policy);
			quit = ret & H_QUIT;
			newpolicy = ret & H_NEWPOLICY;
			polfunc = getpolicyfunc(policy->policy);
		}

		timeout = (*polfunc)(newpolicy, policy);
		newpolicy = 0;
	}

	return 0;
}

int main(int argc, char **argv)
{
	static const struct sockaddr_un addr = { AF_UNIX, SOCKPATH };
	static const int addrlen = offsetof(struct sockaddr_un, sun_path)+sizeof(SOCKPATH)-1;

	int opt;
	int daemonize = 1;
	int usage = 0;
	const char *pidfile = NULL;
	int showversion = 0;
	const char *newpolicy = NULL;
	struct sf_policyinfo policy = { SF_NILPOLICY };
	int ret;

	progname = argv[0];

	while((opt = getopt(argc, argv, "DFP:Vp:u?")) != EOF) {
		switch(opt) {
		case 'p':
			newpolicy = optarg;
			break;

		case 'u':
			clo_userchange = 1;
			break;

		case 'D':
			clo_debug = 1;
			/* FALLTHROUGH */
		case 'F':
			daemonize = 0;
			break;

		case 'P':
			pidfile = optarg;
			break;

		case 'V':
			showversion = 1;
			break;

		case '?':
			usage = 1;
			break;

		default:
			usage = 2;
			break;
		}
	}

	if (optind != argc)
		usage = 2;


	if (newpolicy != NULL) {
		long l;
		char *end;

		if (strcmp(newpolicy, "powersave") == 0)
			policy.policy = SF_POWERSAVE;
		else if (strcmp(newpolicy, "performance") == 0)
			policy.policy = SF_PERFORMANCE;
		else if (strcmp(newpolicy, "dynamic") == 0) {
			policy.policy = SF_DYNAMIC;
			policy.dynamic.min = policy.dynamic.max = 0; /* default */
		} else if ((l = strtol(newpolicy, &end, 10)) && *end == '\0') {
			policy.policy = SF_FIXED;
			policy.fixed.current = (int)l * 1000; /* MHz -> kHz */
		} else {
			fprintf(stderr, "%s: unknown policy \"%s\"\n",
				progname, newpolicy);
			usage = 2;
		}
	}

	if (usage) {
		printf("Usage: %s [-fqsuDPV] [-p policy]\n", progname);
		printf(
			"\t-p P\tset policy to P\n"
			"\t-u\tOK for users to set policy\n"
			"\t-D\tprint debug messages (implies -F)\n"
			"\t-F\tstay in foreground (don't become daemon)\n"
			"\t-P file\tcreate a pid file\n"
			"\t-V\tversion information\n"

			"\nAvailable policies:\n"
			"\tpowersave\tmaximise power savings\n"
			"\tperformance\tmaximise performance\n"
			"\tdynamic\t\tadjust speed according to CPU load (default)\n"
			"\tNNN\t\tset CPU to a fixed speed of NNN MHz\n"
			"\n"
			);
		exit(usage == 2 ? 1 : 0);
	}

	if (geteuid() != 0) {
		fprintf(stderr, "%s: this daemon must be run by root\n", progname);
		exit(1);
	}

	if (!find_cpusys()) {
		fprintf(stderr, "%s: cannot find cpufreq under /sys\n", progname);
		fprintf(stderr, "\tNeed sysfs mounted on /sys, and kernel configured with cpufreq driver\n");
		exit(1);
	}

	/* make sure we can compute idle ratio */
	getidle();

	if (sf_connect() == 0) {
		sf_disconnect();
		fprintf(stderr, "%s: server seems to be already running\n", progname);
		exit(1);
	}

	sockfd = socket(PF_UNIX, SOCK_DGRAM, 0);

	if (sockfd == -1) {
		perror("socket");
		exit(1);
	}

	const int passcred = 1;
	if (setsockopt(sockfd, SOL_SOCKET, SO_PASSCRED, &passcred, sizeof(passcred)) == -1)
		perror("setsockopt/PASSCRED");

	unlink(SOCKPATH);

	if (bind(sockfd, (const struct sockaddr *)&addr, addrlen) == -1) {
		perror("bind - server already running?");
		exit(1);
	}

	if (chmod(SOCKPATH, 0666) == -1)
		perror("chmod");

	if (daemonize && daemon(0, 1) != 0) {
		fprintf(stderr, "%s: daemon failed: %s\n", progname,
			strerror(errno));
		exit(1);
	}

	if (pidfile != NULL) {
		FILE *fp = fopen(pidfile, "w");
		if (fp == NULL) {
			fprintf(stderr, "%s: can't create pidfile %s: %s\n",
				progname, pidfile, strerror(errno));
			exit(1);
		}
		fprintf(fp, "%d\n", getpid());
		fclose(fp);
	}

	if (policy.policy == SF_NILPOLICY) {
		policy.policy = SF_DYNAMIC;
		policy.dynamic.min = policy.dynamic.max = 0;
	}

	/* Get unchanging CPU details */
	readint("cpuinfo_min_freq", &cpuinfo.min);
	readint("cpuinfo_max_freq", &cpuinfo.max);
	if (readstr("scaling_driver", cpuinfo.driver, sizeof(cpuinfo.driver)) != -1) {
		char *cp;

		cpuinfo.driver[sizeof(cpuinfo.driver)-1] = '\0';
		cp = strchr(cpuinfo.driver, '\n');
		if (cp != NULL)
			*cp = '\0';
	} else
		cpuinfo.driver[0] = '\0';

	if (clo_debug) {
		char buf[POLICYLEN];
		policy_str(&policy, buf);
		printf("starting server with policy %s\n",
		       buf);
	}

	/* Run the server */
	ret = server(&policy);

	/* Clean up */
	del_monitorclient(NULL, 0);

	unlink(SOCKPATH);

	if (pidfile != NULL)
		unlink(pidfile);

	return ret;
}
